@@ -2,9 +2,11 @@ use crate::read::BoundedReader;
 use crate::{BookError, Result};
 use aversion::group::{DataSink, DataSourceExt};
 use aversion::util::cbor::CborData;
-use aversion::{assign_message_ids, UpgradeLatest, Versioned};
+use aversion::{assign_message_ids, FromVersion, UpgradeLatest, Versioned};
 use byteorder::{BigEndian, ReadBytesExt, WriteBytesExt};
 use serde::{Deserialize, Serialize};
+use std::convert::TryInto;
+use std::fs::File;
 use std::io::{self, Cursor, Read, Seek, SeekFrom, Write};
 use std::num::NonZeroU64;
 use std::thread::panicking;
@@ -37,7 +39,7 @@ pub type FileHeader = FileHeaderV1;
 /// can be confusing if code attempts to read a zero-sized span. Use
 /// `Option<FileSpan` to represent a zero-sized span.
 ///
-#[derive(Debug, Serialize, Deserialize)]
+#[derive(Debug, PartialEq, Serialize, Deserialize)]
 pub struct FileSpanV1 {
     pub offset: u64,
     pub length: NonZeroU64,
@@ -64,22 +66,51 @@ type FileSpan = FileSpanV1;
 /// This contains an identifying number, and a file span that
 /// tells us what chunk of the file contains this chapter.
 #[derive(Debug, Serialize, Deserialize)]
-pub struct TocEntryV1 {
-    pub id: u64,
+struct TocEntryV1 {
+    id: u64,
+    span: Option<FileSpanV1>,
+}
+
+/// A Table-of-contents entry.
+///
+/// This contains an identifying number, and a file span that
+/// tells us what chunk of the file contains this chapter.
+#[derive(Debug, PartialEq, Serialize, Deserialize)]
+pub struct TocEntryV2 {
+    pub id: Box<[u8]>,
     pub span: Option<FileSpanV1>,
 }
 
 // A type alias, to make code a little easier to read.
-type TocEntry = TocEntryV1;
+type TocEntry = TocEntryV2;
+
+/// A Table-of-contents.
+///
+/// This contains multiple `TocEntry` values, one for each chapter.
+#[derive(Debug, Default, Serialize, Deserialize, Versioned)]
+struct TocV1(Vec<TocEntryV1>);
 
 /// A Table-of-contents.
 ///
 /// This contains multiple `TocEntry` values, one for each chapter.
 #[derive(Debug, Default, Serialize, Deserialize, Versioned, UpgradeLatest)]
-pub struct TocV1(Vec<TocEntryV1>);
+pub struct TocV2(Vec<TocEntryV2>);
+
+impl FromVersion<TocV1> for TocV2 {
+    fn from_version(v1: TocV1) -> Self {
+        let entries =
+            v1.0.into_iter()
+                .map(|v1_entry| TocEntryV2 {
+                    id: Box::new(v1_entry.id.to_be_bytes()),
+                    span: v1_entry.span,
+                })
+                .collect();
+        TocV2(entries)
+    }
+}
 
 // A type alias, used by the Versioned trait.
-type Toc = TocV1;
+type Toc = TocV2;
 
 impl Toc {
     fn add(&mut self, entry: TocEntry) {
@@ -90,8 +121,13 @@ impl Toc {
         self.0.iter()
     }
 
-    fn get_chapter(&self, index: ChapterIndex) -> Result<&TocEntry> {
-        self.0.get(index.0).ok_or(BookError::NoChapter)
+    fn get_chapter<Id>(&self, id: Id) -> Result<&TocEntry>
+    where
+        Id: Into<ChapterId>,
+    {
+        let id: ChapterId = id.into();
+        let entry = self.iter().find(|entry| entry.id == id.0);
+        entry.ok_or(BookError::NoChapter)
     }
 }
 
@@ -100,6 +136,42 @@ assign_message_ids! {
     Toc: 2,
 }
 
+/// A chapter identifier.
+///
+/// This is internally a byte array. Any type may be used as a chapter
+/// identifier, as long as it implements `Into<ChapterId>`.
+pub struct ChapterId(pub Box<[u8]>);
+
+impl From<&[u8]> for ChapterId {
+    fn from(slice: &[u8]) -> Self {
+        Self(slice.into())
+    }
+}
+
+impl From<Box<[u8]>> for ChapterId {
+    fn from(boxed: Box<[u8]>) -> Self {
+        Self(boxed)
+    }
+}
+
+impl From<&str> for ChapterId {
+    fn from(s: &str) -> Self {
+        String::from(s).into()
+    }
+}
+
+impl From<String> for ChapterId {
+    fn from(s: String) -> Self {
+        Self(s.into_bytes().into_boxed_slice())
+    }
+}
+
+impl From<u64> for ChapterId {
+    fn from(n: u64) -> Self {
+        Self(Box::new(n.to_be_bytes()))
+    }
+}
+
 /// A tool for writing a `Chapter`.
 ///
 /// A `ChapterWriter` creates a new chapter. Chapters will be written
@@ -118,7 +190,7 @@ assign_message_ids! {
 /// [`close()`]: Self::close
 pub struct ChapterWriter<W> {
     book: Option<BookWriter<W>>,
-    id: u64,
+    id: Box<[u8]>,
     offset: usize,
     length: usize,
 }
@@ -128,11 +200,15 @@ where
     W: Write,
 {
     /// Create a new `ChapterWriter`.
-    fn new(book: BookWriter<W>, id: u64) -> Self {
+    fn new<Id>(book: BookWriter<W>, id: Id) -> Self
+    where
+        Id: Into<ChapterId>,
+    {
+        let id: ChapterId = id.into();
         let offset = book.current_offset;
         ChapterWriter {
             book: Some(book),
-            id,
+            id: id.0,
             offset,
             length: 0,
         }
@@ -149,7 +225,7 @@ where
         self.flush()?;
 
         let toc_entry = TocEntry {
-            id: self.id,
+            id: self.id.clone(),
             span: FileSpan::from_offset_length(self.offset, self.length),
         };
 
@@ -269,9 +345,13 @@ impl<W: Write> BookWriter<W> {
     /// Create a new `ChapterWriter`.
     ///
     /// The chapter `id` can be any value the user wants, and can be
-    /// used to later locate a chapter.
+    /// used to later locate a chapter. It may be any type that
+    /// implements [`Into<ChapterId>`][ChapterId].
     ///
-    pub fn new_chapter(self, id: u64) -> ChapterWriter<W> {
+    pub fn new_chapter<Id>(self, id: Id) -> ChapterWriter<W>
+    where
+        Id: Into<ChapterId>,
+    {
         ChapterWriter::new(self, id)
     }
 
@@ -301,12 +381,20 @@ impl<W: Write> BookWriter<W> {
     }
 }
 
-/// A chapter index.
-#[derive(Clone, Copy, Debug)]
-pub struct ChapterIndex(pub usize);
-
 /// An interface for reading a Bookfile.
 ///
+/// The `Book` type represents a read-only Bookfile. Invividual chapters can
+/// be read in their entirety, or incrementally using the [`Read`] interface.
+/// [`Seek`] and [`read_at`]/[`read_exact_at`]
+/// are also provided, and work within the context of that chapter: the seek
+/// offset is the offset within the chapter, and a read at the end of the
+/// chapter will return EOF.
+///
+/// [`Read`]: std::io::Read
+/// [`Seek`]: std::io::Seek
+/// [`read_at`]: crate::BoundedReader::read_at
+/// [`read_exact_at`]: crate::BoundedReader::read_exact_at
+
 #[derive(Debug)]
 pub struct Book<R> {
     reader: R,
@@ -315,7 +403,7 @@ pub struct Book<R> {
 }
 
 impl<R> Book<R> {
-    /// Return the file's magic number
+    /// Return the file's magic number.
     ///
     /// Each BookWriter specifies a magic number, used to identify this file format.
     /// This method returns the value found in the file.
@@ -324,11 +412,63 @@ impl<R> Book<R> {
     }
 }
 
+#[cfg(target_family = "unix")]
+impl Book<File> {
+    /// Create a shared reader object.
+    ///
+    /// This returns a BoundedReader that does not implement `Seek` or `Read`, only
+    /// [`read_at`] and [`read_exact_at`]
+    /// This means it can be used via a shared reference.
+    ///
+    /// If the `Seek` and `Read` traits are required, use [`exclusive_chapter_reader`] instead.
+    ///
+    /// [`read_at`]: crate::BoundedReader::read_at
+    /// [`read_exact_at`]: crate::BoundedReader::read_exact_at
+    /// [`exclusive_chapter_reader`]: Self::exclusive_chapter_reader
+    ///
+    pub fn chapter_reader<Id>(&self, index: Id) -> Result<BoundedReader<&File>>
+    where
+        Id: Into<ChapterId>,
+    {
+        let toc_entry = self.toc.get_chapter(index)?;
+        match &toc_entry.span {
+            None => {
+                // If the span is empty, no IO is necessary; just return
+                // an empty Vec.
+                Ok(BoundedReader::empty(&self.reader))
+            }
+            Some(span) => Ok(BoundedReader::new(
+                &self.reader,
+                span.offset,
+                span.length.into(),
+            )),
+        }
+    }
+
+    /// Read all bytes in a chapter.
+    ///
+    /// This is the same thing as calling [`chapter_reader`] followed by
+    /// [`read_exact_at`].
+    ///
+    /// [`chapter_reader`]: Self::chapter_reader
+    /// [`read_exact_at`]: crate::BoundedReader::read_exact_at
+    pub fn read_chapter<Id>(&self, index: Id) -> Result<Box<[u8]>>
+    where
+        Id: Into<ChapterId>,
+    {
+        let reader = self.chapter_reader(index)?;
+        let chapter_len: usize = reader.len().try_into().unwrap();
+        let mut buf = vec![0u8; chapter_len];
+        reader.read_exact_at(&mut buf, 0)?;
+        Ok(buf.into_boxed_slice())
+    }
+}
+
 impl<R> Book<R>
 where
     R: Read + Seek,
 {
-    /// Create a new Book from a stream
+    /// Create a new Book from a stream.
     ///
     /// This call will attempt to read the file header and table of contents.
     /// It may fail due to IO errors while reading, or invalid file data.
@@ -370,22 +510,30 @@ where
         })
     }
 
-    /// Look up a chapter.
+    /// Check whether a chapter exists.
     ///
     /// For now, we assume chapter ids are unique. That's dumb,
     /// and will be fixed in a future version.
-    pub fn find_chapter(&self, id: u64) -> Option<ChapterIndex> {
-        for (index, entry) in self.toc.iter().enumerate() {
-            if entry.id == id {
-                return Some(ChapterIndex(index));
-            }
-        }
-        None
+    pub fn has_chapter<Id>(&self, id: Id) -> bool
+    where
+        Id: Into<ChapterId>,
+    {
+        self.toc.get_chapter(id).is_ok()
     }
 
-    /// Read a chapter by index.
-    pub fn chapter_reader(&mut self, index: ChapterIndex) -> Result<BoundedReader<R>> {
-        let toc_entry = self.toc.get_chapter(index)?;
+    /// Read a chapter, with seeking.
+    ///
+    /// This returns a `BoundedReader` that can read and seek within the chapter
+    /// (assuming the underlying Book implements `Read + Seek`).
+    /// The `Seek` and `Read` traits require exclusive access. For a shared reader,
+    /// use [`chapter_reader`] instead.
+    ///
+    /// [`chapter_reader`]: Self::chapter_reader
+    pub fn exclusive_chapter_reader<Id>(&mut self, id: Id) -> Result<BoundedReader<&mut R>>
+    where
+        Id: Into<ChapterId>,
+    {
+        let toc_entry = self.toc.get_chapter(id)?;
         match &toc_entry.span {
             None => {
                 // If the span is empty, no IO is necessary; just return
@@ -405,13 +553,20 @@ where
 
     /// Read all bytes in a chapter.
     ///
-    /// This is the same thing as calling [`chapter_reader`] followed by
-    /// `read_to_end`.
+    /// This is the same thing as calling [`exclusive_chapter_reader`]
+    /// followed by [`read_to_end`]. It can be used on anything that implements
+    /// `Read + Seek`, but as a result it requires exclusive access via a
+    /// mutable reference.
     ///
-    /// [`chapter_reader`]: Self::chapter_reader
-    pub fn read_chapter(&mut self, index: ChapterIndex) -> Result<Box<[u8]>> {
+    /// [`exclusive_chapter_reader`]: Self::exclusive_chapter_reader
+    ///[`read_to_end`]: std::io::Read::read_to_end
+    ///
+    pub fn exclusive_read_chapter<Id>(&mut self, index: Id) -> Result<Box<[u8]>>
+    where
+        Id: Into<ChapterId>,
+    {
         let mut buf = vec![];
-        let mut reader = self.chapter_reader(index)?;
+        let mut reader = self.exclusive_chapter_reader(index)?;
         reader.read_to_end(&mut buf)?;
         Ok(buf.into_boxed_slice())
     }
@@ -466,24 +621,76 @@ mod tests {
             let mut chapter = book.new_chapter(22);
             chapter.write_all(b"This is chapter 22").unwrap();
             let book = chapter.close().unwrap();
-            let mut chapter = book.new_chapter(33);
+            let mut chapter = book.new_chapter("🦀");
             chapter.write_all(b"This is chapter 33").unwrap();
             let book = chapter.close().unwrap();
             book.close().unwrap()
         };
         let mut book = Book::new(buffer).unwrap();
-        let n = book.find_chapter(11).unwrap();
-        let ch1 = book.read_chapter(n).unwrap();
+        let ch1 = book.exclusive_read_chapter(11).unwrap();
         assert!(ch1.is_empty());
 
-        assert!(book.find_chapter(1).is_none());
+        assert!(!book.has_chapter(1));
 
-        let n = book.find_chapter(22).unwrap();
-        let ch2 = book.read_chapter(n).unwrap();
+        let ch2 = book.exclusive_read_chapter(22).unwrap();
         assert_eq!(ch2.as_ref(), b"This is chapter 22");
 
-        let n = book.find_chapter(33).unwrap();
-        let ch2 = book.read_chapter(n).unwrap();
+        let ch2 = book.exclusive_read_chapter("🦀").unwrap();
         assert_eq!(ch2.as_ref(), b"This is chapter 33");
     }
+
+    #[test]
+    fn book_file_shared() {
+        let temp = tempfile::tempfile().unwrap();
+
+        let magic = 0x1234;
+        let file = {
+            let book = BookWriter::new(temp, magic).unwrap();
+            let chapter = book.new_chapter(11);
+            let book = chapter.close().unwrap();
+            let mut chapter = book.new_chapter(22);
+            chapter.write_all(b"This is chapter 22").unwrap();
+            let book = chapter.close().unwrap();
+            let mut chapter = book.new_chapter("🦀");
+            chapter.write_all(b"This is chapter 33").unwrap();
+            let book = chapter.close().unwrap();
+            book.close().unwrap()
+        };
+        let book = Book::new(file).unwrap();
+        let ch1 = book.read_chapter(11).unwrap();
+        assert!(ch1.is_empty());
+
+        assert!(!book.has_chapter(1));
+
+        let ch2 = book.read_chapter(22).unwrap();
+        assert_eq!(ch2.as_ref(), b"This is chapter 22");
+
+        let ch2 = book.read_chapter("🦀").unwrap();
+        assert_eq!(ch2.as_ref(), b"This is chapter 33");
+    }
+
+    #[test]
+    fn toc_compat() {
+        let mut toc = Vec::new();
+        toc.push(TocEntryV1 {
+            id: 1234,
+            span: Some(FileSpanV1 {
+                length: 33.try_into().unwrap(),
+                offset: 44,
+            }),
+        });
+        let toc = TocV1(toc);
+        let toc = TocV2::from_version(toc);
+        assert_eq!(toc.0.len(), 1);
+        assert_eq!(
+            toc.0[0],
+            TocEntryV2 {
+                id: Box::new(1234u64.to_be_bytes()),
+                span: Some(FileSpanV1 {
+                    length: 33.try_into().unwrap(),
+                    offset: 44,
+                }),
+            }
+        )
+    }
 }
